/*********************************************************************
 * S_Socket.C -- simple AF_INET socket class
 * $Id: S_Socket.C,v 1.12 1998/08/04 15:49:26 tjko Exp $
 *
 * Copyright (c) 1998  Timo Kokkonen <tjko@iki.fi>
 *
 *
 * This file may be copied under the terms and conditions 
 * of the GNU General Public License, as published by the Free
 * Software Foundation (Cambridge, Massachusetts).
 */


#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <netdb.h>
#include <arpa/inet.h>
#ifdef DEBUG
#include <stdarg.h>
#endif

#include "S_Socket.H"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#ifdef BUGGY_GCC
#define GETERRNO errno
#else
#define GETERRNO ::errno
#endif


static char *rcsid = "$Id: S_Socket.C,v 1.12 1998/08/04 15:49:26 tjko Exp $";

#ifdef DEBUG
void S_Socket::debug(char *format, ...)
{
  va_list args;

  fprintf(stderr, "OBJECT(%x) S_Socket::",this);
  va_start(args,format);
  vfprintf(stderr,format,args);
  va_end(args);
  fprintf(stderr,"\n");
  fflush(stderr);
}
 #define DEBUGMSG2(f,a)         debug(f,a)
 #define DEBUGMSG3(f,a,b)       debug(f,a,b)
 #define DEBUGMSG4(f,a,b,c)     debug(f,a,b,c)
 #define DEBUGMSG5(f,a,b,c,d)   debug(f,a,b,c,d)
 #define DEBUGMSG6(f,a,b,c,d,e) debug(f,a,b,c,d,e)
#else
 #define DEBUGMSG2(f,a)
 #define DEBUGMSG3(f,a,b)
 #define DEBUGMSG4(f,a,b,c)
 #define DEBUGMSG5(f,a,b,c,d)
 #define DEBUGMSG6(f,a,b,c,d,e)
#endif




S_Socket::S_Socket() 
{
  if (rcsid);

  fdes = -1;
  err = 0;
  portnum = 0;
  type = 0;
  mode = 0;
  saddr = new in_addr;
  saddress = new sockaddr_in;
  address_str = proto_str = service_str = hostname_str = NULL;
  address_valid = 0;
  port_valid = 0;
  sock_open = 0;
}

S_Socket::~S_Socket()
{
  if (sock_open) ::close(fdes);
  delete saddr;
  delete saddress;
  delete address_str;
  delete proto_str;
  delete service_str;
  delete hostname_str;
}


int S_Socket::setport(char const *service, char const *proto)
{
  struct servent *serv;
  long int lport;

  if (!service || !proto) {
    port_valid=0; err=0;
    return -1;
  }

  /* First try to read it from /etc/services */
  serv = getservbyname(service, proto);
  if (serv) {
    portnum = serv->s_port;
    delete service_str;
    service_str = strdup(serv->s_name);
    delete proto_str;
    proto_str = strdup(serv->s_proto);
    port_valid = 1;
  }
  else { /* Not in services, maybe a number? */
    char *errpos;
    lport = strtol(service,&errpos,0);
    if ( errpos[0] != 0 ) {
      err=0;
      port_valid=0;
      return -1; 
    }
    return setport((int)lport,proto);
  }

  return 0;
}


int S_Socket::setport(int service, char const *proto)
{
  struct servent *serv;
  char buf[256];

  if (!proto) {
    err=0; port_valid=0;
    return -1;
  }

  serv=getservbyport(service,proto);
  if (!serv) { 
    if ( (service<1) || (service>65535) ) { 
      port_valid=0;
      err=GETERRNO; 
      return -1; 
    }
    portnum = htons(service);
    delete service_str;
    sprintf(buf,"%d",service);
    service_str = strdup(buf);
    delete proto_str;
    proto_str = strdup(proto);
    port_valid = 1;
  } else {
    portnum = serv->s_port;
    delete service_str;
    service_str = strdup(serv->s_name);
    delete proto_str;
    proto_str = strdup(serv->s_proto);
    port_valid = 1;
  }

  return 0;
}


int S_Socket::setaddress(char const *address)
{
  hostent *host;

  if (!address || strlen(address)<1) {
    err=0; address_valid=0;
    return -1;
  }

  memset(saddr,0,sizeof(*saddr));


  if (inet_aton(address,saddr)) {
    delete address_str;
    address_str=strdup(inet_ntoa(*saddr));
    delete hostname_str;
    hostname_str=NULL;
    address_valid=1;
    return 0;
  }

  host=gethostbyname(address);
  if (host) {
    *saddr = *((in_addr*)(*host->h_addr_list));
    delete address_str;
    address_str=strdup(inet_ntoa(*saddr));
    delete hostname_str;
    hostname_str = strdup(host->h_name);
    address_valid=1;
    return 0;
  }

  err=0;
  address_valid=0;
  return -1;
}


int S_Socket::setaddress(unsigned int s_addr) 
{ 
  memset(saddr,0,sizeof(*saddr));
  saddr->s_addr=htonl(s_addr); 
  delete address_str;
  address_str=strdup(inet_ntoa(*saddr));
  delete hostname_str;
  hostname_str=NULL;
  address_valid=1;
  return 0; 
}


char const *S_Socket::hostname()
{
  hostent *host;

  if (!address_valid) return NULL;
  if (!hostname_str) {
    host = gethostbyaddr((char*)&saddr->s_addr,sizeof(saddr->s_addr),AF_INET);
    if (host) hostname_str = strdup(host->h_name);
  }
  return hostname_str;
}



int S_Socket::open(int nonblocking)
{
  err=0;

  if (!port_valid || !address_valid) { 
    DEBUGMSG2("open(%d) port or address not valid",nonblocking);
    err=-1; 
    return -1; 
  }

  if (!strcmp(proto_str,"tcp")) type=SOCK_STREAM;
  else if (!strcmp(proto_str,"udp")) type=SOCK_DGRAM;
  else { 
    DEBUGMSG3("open(%d) invalid protocol '%s'",nonblocking,proto_str);
    err=-2; 
    return -1; 
  }

  if (sock_open) close();

  memset((char*)saddress,0,sizeof(*saddress));
  saddress->sin_family = AF_INET;
  saddress->sin_port = portnum;
  saddress->sin_addr.s_addr = saddr->s_addr;

  fdes = socket(AF_INET, type, 0);
  if (fdes<0) { 
    err=GETERRNO; 
    DEBUGMSG3("open(%d) error creating socket: %d",nonblocking,GETERRNO);
    return -1; 
  }

  if (nonblocking) { /* set socket to non-blocking mode */
    int flags = fcntl(fdes,F_GETFL,0);
    if (flags < 0) {
      DEBUGMSG2("open(%d) error in F_GETFL",nonblocking);
      err=GETERRNO;
      return -1;
    }
    if (fcntl(fdes,F_SETFL,flags|O_NONBLOCK)<0) {
      DEBUGMSG2("open(%d) error setting fd to O_NONBLOCK mode",
			   nonblocking);
      err=GETERRNO;
      mode = 0;
    } else 
      mode = O_NONBLOCK;
  } else
    mode = 0;
  
  if (type == SOCK_STREAM) {
    int ret = connect(fdes, (sockaddr *)saddress, sizeof(*saddress));
    DEBUGMSG4("open(%d) 'connect' ret=%d (fd=%d)",nonblocking,ret,fdes);
    if ((ret < 0) && !(nonblocking && (GETERRNO==EINPROGRESS && ret==-1))) { 
      err=GETERRNO; 
      DEBUGMSG2("open() failed. errno=%d",GETERRNO);
      ::close(fdes);
      return -1; 
    }
    sock_open=1;
    return fdes;
  }

  /* must be udp connection so... */
  if (bind(fdes, (sockaddr *)saddress, sizeof(*saddress)) < 0) {
    err=GETERRNO;
    DEBUGMSG3("open(%d) 'bind' err=%d",nonblocking,GETERRNO);
    ::close(fdes);
    return -1;
  }
  DEBUGMSG3("open(%d) 'bind' succesful (fd=%d)",nonblocking,fdes);
  sock_open=1;
  return fdes;
}


void S_Socket::close()
{
  if (!sock_open) {
    DEBUGMSG2("close() tried to close allready closed socket(fd=%d)",fdes);
    return;
  }

  // shutdown(fdes,2);
  ::close(fdes);
  sock_open=0;
  DEBUGMSG2("close() socket closed (fd=%d).",fdes);
}


int S_Socket::read(char *buf, int count)
{
  int this_read, total_read=0;
  err=0; 

  if (!sock_open || !buf || (count<0)) { 
    DEBUGMSG4("read(%x,%d) invalid params %s",buf,count,
	      (!sock_open?", closed socket":""));
    err=-1; 
    return -1; 
  }

  while (total_read < count) {
    do {
      this_read=::read(fdes,buf,count - total_read);
      DEBUGMSG4("read(%x,%d) read ret=%d",buf,count-total_read,this_read);
    } while (this_read<0 && GETERRNO==EINTR);
    
    if (this_read < 0) {
      err=GETERRNO;
      if (GETERRNO==EAGAIN && mode==O_NONBLOCK) return total_read;
      DEBUGMSG2("read() failed errno=%d",GETERRNO);
      return this_read;
    } 
    else if (this_read == 0) return total_read;
    
    total_read += this_read;
    buf += this_read;
  }

  return total_read;
}

char *S_Socket::gets(char *s, int size)
{
  int wasread=1, count = 0;
  char last_read = 0, *pos;

  if (!sock_open || !s || (size<1)) { 
    DEBUGMSG4("gets(%x,%d) invalid params %s",s,size,
	      (!sock_open?", closed socket":""));
    err=-1; 
    return NULL; 
  }

  if (size==1) { *s=0; return s; };

  pos=s;

  do {
    do {
      wasread = ::read(fdes,&last_read,1);
    } while (wasread <0 && GETERRNO==EINTR);

    if (wasread < 0) { 
      err=GETERRNO; 
      return NULL; 
    }
    if (wasread==0) break;

    *(pos++)=last_read;
    count++;

  } while (last_read != 10 && last_read != 0);

  if (count<1) return NULL;

  if (count<size) *(s+count)=0;
  else *(s+size-1)=0;
 
  DEBUGMSG3("gets() succesful '%s' (%d)",s,strlen(s));
  return s;
}


char *S_Socket::getstr(char *s, int size)
{
  char *r;

  if (!sock_open || !s || (size<1)) { err=0; return NULL; }

  if (gets(s,size)==NULL) return NULL;

  r=s+strlen(s)-1;
  while ( (r>s) && ((*r==10)||(*r==13)) ) *(r--)=0;

  return s;
}


int S_Socket::write(char const *buf, int count)
{
  int this_write,bytes_sent=0;

  if (!sock_open || !buf || (count<0)) { 
    DEBUGMSG4("write(%x,%d) invalid params %s",buf,count,
	      (!sock_open?", closed socket":""));
    err=0; 
    return -1; 
  }

  while (bytes_sent < count) {
    do {
      GETERRNO=0;
      this_write=::write(fdes,buf,count-bytes_sent);
      DEBUGMSG5("write(%x,%d) write ret=%d (fd=%d)",buf,count,this_write,fdes);
    } while ( (this_write<0) && (GETERRNO==EINTR) );

    if (this_write < 0) {
      err=GETERRNO;
      DEBUGMSG3("write() failed (%d) errno=%d",this_write,GETERRNO);
      return this_write;
    }
   
    else if (this_write == 0) return bytes_sent;

    bytes_sent+=this_write;
    buf+=this_write;
  }
  
  return bytes_sent;
}


int S_Socket::readyforread(int tmout)
{
  fd_set rfds;
  struct timeval tv;
  int ret;

  if (!sock_open) return -1;

  FD_ZERO(&rfds);
  FD_SET(fdes,&rfds);
  tv.tv_sec = (tmout<0?0:tmout);
  tv.tv_usec = 0;

  ret=select(fdes+1,&rfds,NULL,NULL,&tv);

  if (ret) return 1;
  else return 0;
}


int S_Socket::readyforwrite(int tmout)
{
  fd_set wfds;
  struct timeval tv;
  int ret;

  if (!sock_open) return -1;

  FD_ZERO(&wfds);
  FD_SET(fdes,&wfds);
  tv.tv_sec = (tmout<0?0:tmout);
  tv.tv_usec = 0;

  ret=select(fdes+1,NULL,&wfds,NULL,&tv);

  if (ret) return 1;
  else return 0;
}


//------------------------------------------------------------

void S_Socket::ignore_pipe()
{
  struct sigaction sig;

  sig.sa_handler = SIG_IGN;
  sig.sa_flags = 0;
  sigemptyset(&sig.sa_mask);
  sigaction(SIGPIPE,&sig,NULL);
}
